@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 1993-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 1993-2024 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -23,11 +23,14 @@
 
 #include "gpu/mem_mgr/mem_mgr.h"
 #include "gpu/mem_mgr/heap.h"
+#include "gpu/mem_mgr/phys_mem_allocator/phys_mem_allocator.h"
 #include "gpu/mem_mgr/fbsr.h"
+#include "gpu/mem_sys/kern_mem_sys.h"
 #include "gpu/bus/kern_bus.h"
 #include "vgpu/rpc.h"
 #include "os/nv_memory_type.h"
 #include "gpu/gsp/kernel_gsp.h"
+#include "gpu/gsp/gsp_static_config.h"
 
 static NV_STATUS _memmgrWalkHeap(OBJGPU *pGpu, MemoryManager *pMemoryManager, OBJFBSR *pFbsr);
 static NV_STATUS _memmgrAllocFbsrReservedRanges(OBJGPU *pGpu, MemoryManager *pMemoryManager);
@@ -43,11 +46,16 @@ memmgrSavePowerMgmtState_KERNEL
 )
 {
     NvU32     fbsrStartMode = pMemoryManager->fbsrStartMode;
-    NvBool    bGcOffState   = NV_FALSE;
     NvU32     i;
     OBJFBSR  *pFbsr;
     NV_STATUS rmStatus = NV_OK;
 
+    if (pGpu->getProperty(pGpu, PDB_PROP_GPU_ZERO_FB) ||
+        pGpu->getProperty(pGpu, PDB_PROP_GPU_BROKEN_FB))
+    {
+        return NV_OK;
+    }
+
     if (pMemoryManager->pActiveFbsr)
     {
         NV_PRINTF(LEVEL_ERROR, "!!!!! Calling Suspend on a suspended GPU or ");
@@ -63,12 +71,14 @@ memmgrSavePowerMgmtState_KERNEL
     // might fail is not enough non-paged memory available.
     //
     pMemoryManager->pActiveFbsr = NULL;
+
     //
     // Iterate the heap at top level to avoid inconsistent between
     // several pFbsr instance.
     //
 
-    rmStatus = memmgrAddMemNodes(pGpu, pMemoryManager, bGcOffState);
+    rmStatus = memmgrAddMemNodes(pGpu, pMemoryManager,
+                                 pGpu->getProperty(pGpu, PDB_PROP_GPU_GCOFF_STATE_ENTERING));
 
     if ((rmStatus == NV_OK) &&
         memmgrIsPmaInitialized(pMemoryManager))
@@ -77,13 +87,13 @@ memmgrSavePowerMgmtState_KERNEL
         RANGELISTTYPE  *pSaveList = NULL;
         RANGELISTTYPE  *pSaveCurr;
 
-        if (bGcOffState)
+        if (pGpu->getProperty(pGpu, PDB_PROP_GPU_GCOFF_STATE_ENTERING))
         {
-            rmStatus = pmaBuildAllocatedBlocksList(&pHeap->pmaObject, &pSaveList);
+            rmStatus = pmaBuildAllocatedBlocksList(pHeap->pPmaObject, &pSaveList);
         }
         else
         {
-            rmStatus = pmaBuildPersistentList(&pHeap->pmaObject, &pSaveList);
+            rmStatus = pmaBuildPersistentList(pHeap->pPmaObject, &pSaveList);
         }
 
         if (rmStatus == NV_OK)
@@ -112,13 +122,13 @@ memmgrSavePowerMgmtState_KERNEL
             }
         }
 
-        if (bGcOffState)
+        if (pGpu->getProperty(pGpu, PDB_PROP_GPU_GCOFF_STATE_ENTERING))
         {
-            pmaFreeAllocatedBlocksList(&pHeap->pmaObject, &pSaveList);
+            pmaFreeAllocatedBlocksList(pHeap->pPmaObject, &pSaveList);
         }
         else
         {
-            pmaFreePersistentList(&pHeap->pmaObject, &pSaveList);
+            pmaFreePersistentList(pHeap->pPmaObject, &pSaveList);
         }
     }
 
@@ -212,6 +222,12 @@ memmgrRestorePowerMgmtState_KERNEL
     NV_STATUS    status        = NV_OK;
     NvBool       bIsGpuLost    = NV_FALSE;
 
+    if (pGpu->getProperty(pGpu, PDB_PROP_GPU_ZERO_FB) ||
+        pGpu->getProperty(pGpu, PDB_PROP_GPU_BROKEN_FB))
+    {
+        return NV_OK;
+    }
+
     if (portAtomicIncrementS32(&concurrentfbRestorePowerMgmtStateAccess[pGpu->gpuInstance]) != 1)
     {
         NV_PRINTF(LEVEL_INFO, "Concurrent access");
@@ -253,31 +269,6 @@ done:
     return status;
 }
 
-/*
- * FB regions with LostOnSuspend set to NV_TRUE are expected to be grouped together towards the end of FB
- * First region to have this property set would be the end of save area for Suspend.
- * TODO: Handle LostOnSuspend for non-consequetive regions
- */
-static NvU64
-_memmgrGetFbEndExcludingLostOnSuspendRegions
-(
-    OBJGPU        *pGpu,
-    MemoryManager *pMemoryManager
-)
-{
-    NvU32 i = 0;
-
-    for (i = 0; i < pMemoryManager->Ram.numFBRegions; i++)
-    {
-        if (pMemoryManager->Ram.fbRegion[i].bLostOnSuspend == NV_TRUE)
-        {
-            return pMemoryManager->Ram.fbRegion[i].base;
-        }
-    }
-
-    return (pMemoryManager->Ram.fbAddrSpaceSizeMb << 20);
-}
-
 static NV_STATUS
 _memmgrAllocFbsrReservedRanges
 (
@@ -285,15 +276,18 @@ _memmgrAllocFbsrReservedRanges
     MemoryManager *pMemoryManager
 )
 {
-    NvU64      size       = 0;
     NV_STATUS  status     = NV_OK;
     KernelBus *pKernelBus = GPU_GET_KERNEL_BUS(pGpu);
+    NvU64      rsvdMemEnd = pMemoryManager->rsvdMemoryBase + pMemoryManager->rsvdMemorySize;
+    // We need to skip BAR2 CPU visible range during restore, as it needs to be up before allocations are restored
+    NvU64 bar2CpuVisiblePteStart = pKernelBus->bar2[GPU_GFID_PF].pteBase;
+    NvU64 bar2CpuVisiblePteEnd   = bar2CpuVisiblePteStart + pKernelBus->bar2[GPU_GFID_PF].cpuVisiblePgTblSize;
 
     // Alloc the Memory descriptors for Fbsr Reserved regions, if not allocated.
     if (pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_BEFORE_BAR2PTE] == NULL)
     {
-        // Allocate Vid Mem descriptor for RM INSTANCE memory from start to BAR2PTE
-        size = memdescGetPhysAddr(pKernelBus->virtualBar2[GPU_GFID_PF].pPTEMemDesc, AT_GPU, 0) - pMemoryManager->rsvdMemoryBase;
+        NvU64 size = bar2CpuVisiblePteStart - pMemoryManager->rsvdMemoryBase;
+
         NV_ASSERT_OK_OR_GOTO(status,
                              memdescCreate(&pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_BEFORE_BAR2PTE],
                                             pGpu, size, 0, NV_TRUE, ADDR_FBMEM,
@@ -306,73 +300,78 @@ _memmgrAllocFbsrReservedRanges
 
     if (pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_AFTER_BAR2PTE] == NULL)
     {
-        // Allocate Mem descriptors for AFTER_BAR2PTE, GSP HEAP, WPR, NON WPR and VGA Workspace regions
+        NvU64 size = rsvdMemEnd - bar2CpuVisiblePteEnd;
+
+        NV_ASSERT_OK_OR_GOTO(status,
+                             memdescCreate(&pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_AFTER_BAR2PTE],
+                                           pGpu, size, 0, NV_TRUE, ADDR_FBMEM,
+                                           NV_MEMORY_UNCACHED, MEMDESC_FLAGS_NONE),
+                             fail);
+        memdescDescribe(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_AFTER_BAR2PTE],
+                        ADDR_FBMEM,
+                        bar2CpuVisiblePteEnd, size);
+    }
+
+    if (pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_HEAP] == NULL)
+    {
+        /*
+         * Allocate memdesc for AFTER_BAR2PTE and GSP WPR regions.
+         * Allocate sysmem to save GSP allocations
+         */
         if (IS_GSP_CLIENT(pGpu))
         {
-            KernelGsp *pKernelGsp              = GPU_GET_KERNEL_GSP(pGpu);
-            RmPhysAddr afterBar2PteRegionStart = memdescGetPhysAddr(pKernelBus->virtualBar2[GPU_GFID_PF].pPTEMemDesc, AT_GPU, 0) +
-                                                 pKernelBus->virtualBar2[GPU_GFID_PF].pPTEMemDesc->Size;
-            NvU64      afterBar2PteRegionEnd   = pMemoryManager->rsvdMemoryBase + pMemoryManager->rsvdMemorySize;
-            NvU64      afterBar2PteRegionSize  = afterBar2PteRegionEnd - afterBar2PteRegionStart;
-            NvU64      gspHeapRegionStart      = afterBar2PteRegionEnd;
-            NvU64      gspHeapRegionSize       = pKernelGsp->pWprMeta->gspFwRsvdStart - gspHeapRegionStart;
-            NvU64      gspNonWprRegionSize     = pKernelGsp->pWprMeta->gspFwWprStart  - pKernelGsp->pWprMeta->gspFwRsvdStart;
+            KernelGsp    *pKernelGsp = GPU_GET_KERNEL_GSP(pGpu);
+            GspFwWprMeta *pWprMeta   = pKernelGsp->pWprMeta;
 
-            // Create memdesc of AFTER BAR2PTE, GSP HEAP, GSP NON WPR and VGA Workspace regions
-            NV_ASSERT_OK_OR_GOTO(status,
-                                 memdescCreate(&pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_AFTER_BAR2PTE],
-                                                pGpu, afterBar2PteRegionSize, 0, NV_TRUE,
-                                                ADDR_FBMEM, NV_MEMORY_UNCACHED, MEMDESC_FLAGS_NONE),
-                                 fail);
+            /*
+             * Calculate sysmem required to save all GSP allocations
+             * This is sum of GSP Heap, GSP non WPR and VGA workspace regions
+             * This will also include CBC region if the corresponding flag is set
+             *
+             * SYSMEM required for GSP allocations will be created and allocated
+             * into fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_HEAP]
+             *
+             * TODO: Query GSP for size of its allocation instead of this calculation
+             */
+            NvU64 size = ((pWprMeta->nonWprHeapOffset + pWprMeta->nonWprHeapSize) - rsvdMemEnd) + // GSP Heap start to end of NON WPR region
+                         pWprMeta->vgaWorkspaceSize;                                              // VGA Workspace
+
+            // Check if CBC region needs to be saved
+            if (GPU_GET_KERNEL_MEMORY_SYSTEM(pGpu)->bPreserveComptagBackingStoreOnSuspend ||
+                pGpu->getProperty(pGpu, PDB_PROP_GPU_RTD3_GCOFF_SUPPORTED))
+            {
+                NV0080_CTRL_FB_GET_COMPBIT_STORE_INFO_PARAMS compbitStoreInfoParams;
+                RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);
+
+                // GET CBC region info from GSP
+                NV_ASSERT_OK_OR_GOTO(status,
+                                     pRmApi->Control(pRmApi,
+                                                     pGpu->hInternalClient,
+                                                     pGpu->hInternalDevice,
+                                                     NV0080_CTRL_CMD_FB_GET_COMPBIT_STORE_INFO,
+                                                     &compbitStoreInfoParams,
+                                                     sizeof(compbitStoreInfoParams)),
+                                     fail);
+
+                size += compbitStoreInfoParams.Size;
+            }
+
+            // Create memdesc to save GSP allocations
             NV_ASSERT_OK_OR_GOTO(status,
                                  memdescCreate(&pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_HEAP],
-                                                pGpu, gspHeapRegionSize, 0, NV_TRUE,
-                                                ADDR_FBMEM, NV_MEMORY_UNCACHED, MEMDESC_FLAGS_NONE),
-                                 fail);
-            NV_ASSERT_OK_OR_GOTO(status,
-                                 memdescCreate(&pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_NON_WPR],
-                                                pGpu, gspNonWprRegionSize, 0, NV_TRUE,
-                                                ADDR_FBMEM, NV_MEMORY_UNCACHED, MEMDESC_FLAGS_NONE),
-                                 fail);
-            NV_ASSERT_OK_OR_GOTO(status,
-                                 memdescCreate(&pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_VGA_WORKSPACE],
-                                                pGpu, pKernelGsp->pWprMeta->vgaWorkspaceSize, 0, NV_TRUE,
-                                                ADDR_FBMEM, NV_MEMORY_UNCACHED, MEMDESC_FLAGS_NONE),
+                                                pGpu, size, 0, NV_FALSE,
+                                                ADDR_SYSMEM, NV_MEMORY_UNCACHED, MEMDESC_FLAGS_NONE),
                                  fail);
 
-            // Describe memdesc of AFTER BAR2PTE, GSP HEAP, GSP NON WPR and VGA Workspace regions
-            memdescDescribe(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_AFTER_BAR2PTE],
-                            ADDR_FBMEM,
-                            afterBar2PteRegionStart,
-                            afterBar2PteRegionSize);
-            memdescDescribe(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_HEAP],
-                            ADDR_FBMEM,
-                            gspHeapRegionStart,
-                            gspHeapRegionSize);
-            memdescDescribe(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_NON_WPR],
-                            ADDR_FBMEM,
-                            pKernelGsp->pWprMeta->gspFwRsvdStart,
-                            gspNonWprRegionSize);
-            memdescDescribe(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_VGA_WORKSPACE],
-                            ADDR_FBMEM,
-                            pKernelGsp->pWprMeta->vgaWorkspaceOffset,
-                            pKernelGsp->pWprMeta->vgaWorkspaceSize);
-
-        }
-        // Allocate Vid Mem descriptor for RM INSTANCE memory, specific to VGA  i.e. after BAR2PTE to end.
-        else
-        {
-            NvU64 fbAddrSpaceSize = _memmgrGetFbEndExcludingLostOnSuspendRegions(pGpu, pMemoryManager);
-            size = (fbAddrSpaceSize) - memdescGetPhysAddr(pKernelBus->virtualBar2[GPU_GFID_PF].pPTEMemDesc, AT_GPU, 0) - pKernelBus->virtualBar2[GPU_GFID_PF].pPTEMemDesc->Size;
-
-            NV_ASSERT_OK_OR_GOTO(status,
-                                 memdescCreate(&pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_AFTER_BAR2PTE],
-                                               pGpu, size, 0, NV_TRUE, ADDR_FBMEM,
-                                               NV_MEMORY_UNCACHED, MEMDESC_FLAGS_NONE),
-                                 fail);
-            memdescDescribe(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_AFTER_BAR2PTE],
-                            ADDR_FBMEM,
-                            memdescGetPhysAddr(pKernelBus->virtualBar2[GPU_GFID_PF].pPTEMemDesc, AT_GPU, 0) + pKernelBus->virtualBar2[GPU_GFID_PF].pPTEMemDesc->Size, size);
+            // Allocate sysmem
+            status = memdescAlloc(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_HEAP]);
+            if (status != NV_OK)
+            {
+                NV_ASSERT(status == NV_OK);
+                memdescDestroy(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_HEAP]);
+                pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_HEAP] = NULL;
+                goto fail;
+            }
 
         }
     }
@@ -391,28 +390,16 @@ fail:
     if (pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_HEAP] != NULL)
         memdescDestroy(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_HEAP]);
 
-    if (pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_NON_WPR] != NULL)
-        memdescDestroy(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_NON_WPR]);
-
-    if (pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_WPR] != NULL)
-        memdescDestroy(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_WPR]);
-
-    if (pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_VGA_WORKSPACE] != NULL)
-        memdescDestroy(pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_VGA_WORKSPACE]);
-
     pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_BEFORE_BAR2PTE] = NULL;
     pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_AFTER_BAR2PTE]  = NULL;
     pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_HEAP]       = NULL;
-    pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_NON_WPR]    = NULL;
-    pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_WPR]        = NULL;
-    pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_VGA_WORKSPACE]  = NULL;
 
     return status;
 }
 
 /*!
  * Accessor for list of video memory regions that need to be preserved. This
- * routine calls POBJFBSR::fbsrCopyMemory[|MemDesc] per-region.
+ * routine calls fbsrCopyMemory[|MemDesc] per-region.
  *
  * @param[in]   pFbsr     OBJFBSR pointer
  *
@@ -447,45 +434,6 @@ _memmgrWalkHeap(OBJGPU *pGpu, MemoryManager *pMemoryManager, OBJFBSR *pFbsr)
                                   pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_BEFORE_BAR2PTE]);
         fbsrCopyMemoryMemDesc_HAL(pGpu, pFbsr,
                                   pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_AFTER_BAR2PTE]);
-
-        if (IS_GSP_CLIENT(pGpu))
-        {
-            //
-            // FBSR_OP_SIZE_BUF increments pFbsr->numRegions and updates pFbsr->Length to add sysmem for allocations
-            // FBSR_OP_SAVE sends memdesc info to GSP for CE save
-            //
-            // Handle both FBSR_OP_SIZE_BUF and FBSR_OP_SAVE for GSP NON WPR regions
-            // to allocate sysmem for the regions and to use CE to save/restore the region
-            //
-            fbsrCopyMemoryMemDesc_HAL(pGpu, pFbsr,
-                                      pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_NON_WPR]);
-
-            //
-            // Handle only FBSR_OP_SIZE_BUF for GSP HEAP, GSP WPR and VGA workspace regions
-            // to allocate sysmem for these regions
-            // Use Bus copy to save/restore the GSP WPR region if allocated
-            //
-            if (pFbsr->op == FBSR_OP_SIZE_BUF)
-            {
-                // SysOffset for GSP memory allocations is the AFTER BAR2PTE region's sysoffset
-                pFbsr->gspFbAllocsSysOffset = pFbsr->length;
-
-                fbsrCopyMemoryMemDesc_HAL(pGpu, pFbsr,
-                                          pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_HEAP]);
-                fbsrCopyMemoryMemDesc_HAL(pGpu, pFbsr,
-                                          pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_VGA_WORKSPACE]);
-                pFbsr->numRegions -= 2;
-
-                // GSP WPR memdesc is not NULL when using unauthenticated GSP boot and bus copy
-                if (pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_WPR] != NULL)
-                {
-                    fbsrCopyMemoryMemDesc_HAL(pGpu, pFbsr,
-                                              pMemoryManager->fbsrReservedRanges[FBSR_RESERVED_INST_MEMORY_GSP_WPR]);
-                    pFbsr->numRegions--;
-                }
-            }
-
-        }
     }
     else
     {
@@ -522,7 +470,7 @@ done:
  *              Error otherwise.
  */
 NV_STATUS
-memmgrAddMemNode
+memmgrAddMemNode_IMPL
 (
     OBJGPU            *pGpu,
     MemoryManager     *pMemoryManager,
@@ -571,17 +519,24 @@ memmgrAddMemNode
 }
 
 NV_STATUS
-memmgrAddMemNodes_IMPL(OBJGPU *pGpu, MemoryManager *pMemoryManager, NvBool bSaveAllRmAllocations)
+memmgrAddMemNodes_IMPL
+(
+    OBJGPU        *pGpu,
+    MemoryManager *pMemoryManager,
+    NvBool         bSaveAllRmAllocations
+)
 {
     NV_STATUS           status    = NV_OK;
-    Heap               *pHeap       = GPU_GET_HEAP(pGpu);
+    Heap               *pHeap     = GPU_GET_HEAP(pGpu);
     MEMORY_DESCRIPTOR  *pAllocMemDesc;
     MEM_BLOCK          *block;
+    NvBool              bSaveNode = NV_FALSE;
 
     block = pHeap->pBlockList;
     do
     {
         pAllocMemDesc = block->pMemDesc;
+        bSaveNode     = NV_FALSE;
 
         //
         // TODO: Bug 1778161: Let memory descriptor lost on suspend be default,
@@ -589,18 +544,36 @@ memmgrAddMemNodes_IMPL(OBJGPU *pGpu, MemoryManager *pMemoryManager, NvBool bSave
         // content of memory descriptor, remove MEMDESC_FLAGS_LOST_ON_SUSPEND
         // flag.
         //
-        if ((pAllocMemDesc != NULL) &&
-            (((block->owner == HEAP_OWNER_RM_RESERVED_REGION) &&
-              (bSaveAllRmAllocations ||
-               memdescGetFlag(pAllocMemDesc, MEMDESC_FLAGS_PRESERVE_CONTENT_ON_SUSPEND))) ||
-             (((block->owner == HEAP_OWNER_RM_CHANNEL_CTX_BUFFER) ||
-               (block->owner == HEAP_OWNER_RM_KERNEL_CLIENT)) &&
-              (bSaveAllRmAllocations ||
-               (!memdescGetFlag(pAllocMemDesc, MEMDESC_FLAGS_LOST_ON_SUSPEND))))))
+        if (block->owner != NVOS32_BLOCK_TYPE_FREE && pAllocMemDesc != NULL)
         {
-            if ((memdescGetAddressSpace(pAllocMemDesc) == ADDR_FBMEM) ||
-                (memdescGetAddressSpace(pAllocMemDesc) == ADDR_SYSMEM))
+            NV_PRINTF(LEVEL_INFO,
+                      "pAllocMemDesc base 0x%llx (size 0x%llx) block owner 0x%X memdesc flags 0x%llx\n",
+                      memdescGetPhysAddr(pAllocMemDesc, AT_GPU, 0),
+                      pAllocMemDesc->Size,
+                      block->owner,
+                      pAllocMemDesc->_flags);
+
+            if (memdescGetFlag(pAllocMemDesc, MEMDESC_FLAGS_PRESERVE_CONTENT_ON_SUSPEND))
             {
+                bSaveNode = NV_TRUE;
+            }
+            else if  (block->owner == HEAP_OWNER_RM_RESERVED_REGION)
+            {
+                if (bSaveAllRmAllocations && (!memdescGetFlag(pAllocMemDesc, MEMDESC_FLAGS_LOST_ON_SUSPEND)))
+                    bSaveNode = NV_TRUE;
+            }
+            else if ((block->owner == HEAP_OWNER_RM_CHANNEL_CTX_BUFFER) ||
+                     (block->owner == HEAP_OWNER_RM_KERNEL_CLIENT))
+            {
+                if (bSaveAllRmAllocations || (!memdescGetFlag(pAllocMemDesc, MEMDESC_FLAGS_LOST_ON_SUSPEND)))
+                    bSaveNode = NV_TRUE;
+            }
+
+            if (bSaveNode &&
+                ((memdescGetAddressSpace(pAllocMemDesc) == ADDR_FBMEM) ||
+                 (memdescGetAddressSpace(pAllocMemDesc) == ADDR_SYSMEM)))
+            {
+                NV_PRINTF(LEVEL_INFO, "pAllocMemDesc being saved\n");
                 NV_CHECK_OK_OR_GOTO(status,
                                     LEVEL_ERROR,
                                     memmgrAddMemNode(pGpu, pMemoryManager, pAllocMemDesc, NV_FALSE),
