@@ -1,5 +1,5 @@
 /*
- * SPDX-FileCopyrightText: Copyright (c) 1993-2022 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
+ * SPDX-FileCopyrightText: Copyright (c) 1993-2025 NVIDIA CORPORATION & AFFILIATES. All rights reserved.
  * SPDX-License-Identifier: MIT
  *
  * Permission is hereby granted, free of charge, to any person obtaining a
@@ -26,12 +26,13 @@
 #include "gpu/mem_mgr/mem_mgr.h"
 #include "virtualization/hypervisor/hypervisor.h"
 #include "vgpu/vgpu_events.h"
-#include "objrpc.h"
+#include "gpu/rpc/objrpc.h"
 #include "gpu/bif/kernel_bif.h"
 #include "gpu/bus/kern_bus.h"
-#include "os/os.h"
-#include "nvRmReg.h"
+#include "platform/sli/sli.h"
+#include "nvrm_registry.h"
 #include "gpu/gsp/gsp_static_config.h"
+#include "ctrl/ctrl2080/ctrl2080pmgr.h" // NV2080_CTRL_PMGR_MODULE_INFO_NVSWITCH_SUPPORTED
 
 static void
 kmemsysInitRegistryOverrides
@@ -50,6 +51,24 @@ kmemsysInitRegistryOverrides
         if (data32 == NV_REG_STR_RM_L2_CLEAN_FB_PULL_DISABLED)
             pKernelMemorySystem->bL2CleanFbPull = NV_FALSE;
     }
+
+    if ((osReadRegistryDword(pGpu, NV_REG_STR_RM_OVERRIDE_TO_GMK, &data32) == NV_OK) &&
+        (data32 != NV_REG_STR_RM_OVERRIDE_TO_GMK_DISABLED))
+    {
+        pKernelMemorySystem->overrideToGMK = data32;
+    }
+
+    if ((osReadRegistryDword(pGpu, NV_REG_STR_RM_OVERRIDE_COHERENT_CPU_FB_BASE, &data32)) == NV_OK)
+    {
+        pKernelMemorySystem->coherentCpuFbBaseOverride.bEnabled = NV_TRUE;
+        pKernelMemorySystem->coherentCpuFbBaseOverride.value = (NvU64)data32 << 32ULL;
+    }
+
+    if ((osReadRegistryDword(pGpu, NV_REG_STR_RM_OVERRIDE_NON_PASID_ATS_SUPPORT, &data32)) == NV_OK)
+    {
+        pKernelMemorySystem->nonPasIdAtsOverride.bEnabled = NV_TRUE;
+        pKernelMemorySystem->nonPasIdAtsOverride.bValue = !!data32;
+    }
 }
 
 NV_STATUS
@@ -71,11 +90,50 @@ kmemsysConstructEngine_IMPL
         // resetting GPU FALCONs and in particular resetting the PMU as part of VBIOS
         // init.
         NV_ASSERT_OK_OR_RETURN(kmemsysInitFlushSysmemBuffer_HAL(pGpu, pKernelMemorySystem));
+
     }
 
     return NV_OK;
 }
 
+NV_STATUS
+kmemsysStatePreInitLocked_IMPL
+(
+    OBJGPU *pGpu,
+    KernelMemorySystem *pKernelMemorySystem
+)
+{
+    NV_STATUS status = NV_OK;
+    MEMORY_SYSTEM_STATIC_CONFIG *pStaticConfig;
+
+    pStaticConfig =
+        portMemAllocNonPaged(sizeof(*pStaticConfig));
+    NV_ASSERT_TRUE_OR_GOTO(status,
+        (pStaticConfig != NULL),
+        NV_ERR_INSUFFICIENT_RESOURCES,
+        exit);
+    portMemSet(pStaticConfig, 0, sizeof(*pStaticConfig));
+    //
+    // Assign the pointer immediately after allocating to make sure it will
+    // always be freed
+    //
+    pKernelMemorySystem->pStaticConfig = pStaticConfig;
+
+    NV_ASSERT_OK_OR_GOTO(status,
+        kmemsysInitStaticConfig_HAL(
+            pGpu, pKernelMemorySystem, pStaticConfig),
+        exit);
+
+exit:
+    if (status != NV_OK)
+    {
+        portMemFree((void *)pKernelMemorySystem->pStaticConfig);
+        pKernelMemorySystem->pStaticConfig = NULL;
+    }
+
+    return status;
+}
+
 /*
  * Initialize the Kernel Memory System state.
  *
@@ -90,21 +148,11 @@ NV_STATUS kmemsysStateInitLocked_IMPL
     KernelMemorySystem *pKernelMemorySystem
 )
 {
-    MEMORY_SYSTEM_STATIC_CONFIG *pStaticConfig;
     NV_STATUS status = NV_OK;
+    KernelBus *pKernelBus = GPU_GET_KERNEL_BUS(pGpu);
 
     NV_ASSERT_OK_OR_GOTO(status, kmemsysEnsureSysmemFlushBufferInitialized(pGpu, pKernelMemorySystem), fail);
 
-    pStaticConfig = portMemAllocNonPaged(sizeof(*pStaticConfig));
-    NV_CHECK_OR_RETURN(LEVEL_ERROR, pStaticConfig != NULL, NV_ERR_INSUFFICIENT_RESOURCES);
-    portMemSet(pStaticConfig, 0, sizeof(*pStaticConfig));
-
-    NV_CHECK_OK_OR_GOTO(status, LEVEL_ERROR,
-        kmemsysInitStaticConfig_HAL(pGpu, pKernelMemorySystem, pStaticConfig),
-        fail);
-
-    pKernelMemorySystem->pStaticConfig = pStaticConfig;
-
     KernelBif *pKernelBif = GPU_GET_KERNEL_BIF(pGpu);
 
     pKernelMemorySystem->memPartitionNumaInfo = portMemAllocNonPaged(sizeof(MEM_PARTITION_NUMA_INFO) * KMIGMGR_MAX_GPU_SWIZZID);
@@ -138,11 +186,13 @@ NV_STATUS kmemsysStateInitLocked_IMPL
 
                 pGpu->setProperty(pGpu, PDB_PROP_GPU_ATS_SUPPORTED, NV_TRUE);
             }
-
+        }
+        if (IS_GSP_CLIENT(pGpu) || IS_VIRTUAL_WITH_SRIOV(pGpu))
+        {
             //
-            // PDB_PROP_GPU_C2C_SYSMEM is already set in physical RM but not in
-            // in Kernel-RM where it is actually consumed. setting PDB_PROP_GPU_C2C_SYSMEM
-            // in Kernel-RM when the platform is self-hosted and the C2C links are up, which
+            // PDB_PROP_GPU_C2C_SYSMEM is already set in physical-RM but not in
+            // in Kernel-RM/Guest-RM where it is actually consumed. setting PDB_PROP_GPU_C2C_SYSMEM
+            // in Kernel-RM/Guest-RM when the platform is self-hosted and the C2C links are up, which
             // indicate the C2C is connected to CPU and Physical-RM would have set up the HSHUB
             // to route sysmem through C2C.
             //
@@ -155,14 +205,31 @@ NV_STATUS kmemsysStateInitLocked_IMPL
         // KernelMemorySystem for C2C, NUMA functionality.
         //
         NV_ASSERT_OK_OR_GOTO(status, kmemsysSetupCoherentCpuLink(pGpu, pKernelMemorySystem, NV_FALSE), fail);
+
+        if (gpuIsSelfHosted(pGpu) && pGpu->getProperty(pGpu, PDB_PROP_GPU_COHERENT_CPU_MAPPING))
+        {
+            //Disable BAR1 only for Hopper as Blackwell+ supports simultaneous BAR1 access to FB.
+            if (IsHOPPER(pGpu))
+            {
+                pKernelBus->bBar1Disabled = NV_TRUE;
+            }
+            // In self-hosted systems Disable BAR2 when C2C is up.
+            pKernelBus->bCpuVisibleBar2Disabled = NV_TRUE;
+        }
+    }
+
+    {
+        KernelGmmu   *pKernelGmmu   = GPU_GET_KERNEL_GMMU(pGpu);
+
+        //
+        // Ask GMMU to set the large page size after we have initialized
+        // memory and before we initialize BAR2.
+        //
+        NV_CHECK_OK_OR_RETURN(LEVEL_ERROR,
+            kgmmuCheckAndDecideBigPageSize_HAL(pGpu, pKernelGmmu));
     }
 
 fail:
-    if (status != NV_OK)
-    {
-        portMemFree((void *)pKernelMemorySystem->pStaticConfig);
-    }
-
     return status;
 }
 
@@ -180,7 +247,7 @@ kmemsysStatePreLoad_IMPL
     // to ucode won't program the register itself but will assert that its contents are valid).
     //
     kmemsysProgramSysmemFlushBuffer_HAL(pGpu, pKernelMemorySystem);
-    kmemsysAssertSysmemFlushBufferValid_HAL(pGpu, pKernelMemorySystem);
+    NV_ASSERT_OK_OR_RETURN(kmemsysAssertSysmemFlushBufferValid_HAL(pGpu, pKernelMemorySystem));
 
     // Self Hosted GPUs should have its memory onlined by now.
     if (gpuIsSelfHosted(pGpu) &&
@@ -208,9 +275,28 @@ kmemsysStatePostLoad_IMPL
 )
 {
     if (IS_SILICON(pGpu) &&
-        pGpu->getProperty(pGpu, PDB_PROP_GPU_ATS_SUPPORTED))
+        osNumaOnliningEnabled(pGpu->pOsGpuInfo))
+    {
+        NV_ASSERT_OR_RETURN(pGpu->getProperty(pGpu, PDB_PROP_GPU_ATS_SUPPORTED),
+                            NV_ERR_INVALID_STATE);
+    }
+
+    //
+    // VBIOS programs the peer ATS range register in Nvswitch case.
+    //
+    // For the SRIOV vGPU, the Guest RM chooses the peer id. Programming
+    // by default on the host could potentially create a wrong peer id
+    // being set as host has differnt understanding of the peer value
+    // than the guest. So skip for vGPU host.
+    //
+
+    if (IS_SILICON(pGpu) && !hypervisorIsVgxHyper() &&
+        pGpu->getProperty(pGpu, PDB_PROP_GPU_ATS_SUPPORTED) &&
+        GPU_GET_KERNEL_NVLINK(pGpu) != NULL &&
+        !GPU_IS_NVSWITCH_DETECTED(pGpu))
     {
         NV_STATUS status = kmemsysSetupAllAtsPeers_HAL(pGpu, pKernelMemorySystem);
+
         if (status != NV_OK)
         {
             NV_PRINTF(LEVEL_ERROR, "ATS peer setup failed.\n");
@@ -229,8 +315,9 @@ kmemsysStatePreUnload_IMPL
     NvU32 flags
 )
 {
-    if (IS_SILICON(pGpu) &&
-        pGpu->getProperty(pGpu, PDB_PROP_GPU_ATS_SUPPORTED))
+    if (IS_SILICON(pGpu) && !hypervisorIsVgxHyper() &&
+        pGpu->getProperty(pGpu, PDB_PROP_GPU_ATS_SUPPORTED) &&
+        !GPU_IS_NVSWITCH_DETECTED(pGpu))
     {
         kmemsysRemoveAllAtsPeers_HAL(pGpu, pKernelMemorySystem);
     }
@@ -259,6 +346,7 @@ void kmemsysStateDestroy_IMPL
     }
 
     portMemFree((void *)pKernelMemorySystem->pStaticConfig);
+
 }
 
 /*!
@@ -342,6 +430,7 @@ kmemsysAllocComprResources_KERNEL
     if (!memmgrIsScrubOnFreeEnabled(pMemoryManager))
     {
         if (!(IS_SIMULATION(pGpu) || IsDFPGA(pGpu) || (IS_EMULATION(pGpu) && RMCFG_FEATURE_PLATFORM_MODS)
+            ||(pGpu->getProperty(pGpu, PDB_PROP_GPU_ZERO_FB))
             ||(RMCFG_FEATURE_PLATFORM_WINDOWS && !pGpu->getProperty(pGpu, PDB_PROP_GPU_IN_TCC_MODE))
             ||hypervisorIsVgxHyper()
             ||IS_GFID_VF(gfid)
@@ -382,12 +471,10 @@ kmemsysInitStaticConfig_KERNEL
 )
 {
     RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);
-    NV_STATUS status;
 
-    status = pRmApi->Control(pRmApi, pGpu->hInternalClient, pGpu->hInternalSubdevice,
-                                NV2080_CTRL_CMD_INTERNAL_MEMSYS_GET_STATIC_CONFIG,
-                                pConfig, sizeof(*pConfig));
-    return status;
+    return pRmApi->Control(pRmApi, pGpu->hInternalClient, pGpu->hInternalSubdevice,
+                           NV2080_CTRL_CMD_INTERNAL_MEMSYS_GET_STATIC_CONFIG,
+                           pConfig, sizeof(*pConfig));
 }
 
 /*!
@@ -691,13 +778,13 @@ kmemsysPopulateMIGGPUInstanceMemConfig_KERNEL
         }
 
         //
-        // In GH180 for all the swizzId's for a given memory profile (FULL, HALF, QUARTER 
+        // In GH180 for all the swizzId's for a given memory profile (FULL, HALF, QUARTER
         // and EIGHTH partitions) might not be same. Modify numaMigPartitionSize array
         // for the partition size to be constant for a given profile. BUG 4284299.
         //
         for (memSize = NV2080_CTRL_GPU_PARTITION_FLAG_MEMORY_SIZE_FULL; memSize < NV2080_CTRL_GPU_PARTITION_FLAG_MEMORY_SIZE__SIZE; memSize++)
         {
-            NV_RANGE swizzRange = kmigmgrMemSizeFlagToSwizzIdRange(pGpu, pKernelMIGManager, 
+            NV_RANGE swizzRange = kmigmgrMemSizeFlagToSwizzIdRange(pGpu, pKernelMIGManager,
                                       DRF_NUM(2080_CTRL_GPU, _PARTITION_FLAG, _MEMORY_SIZE, memSize));
             _kmemsysSetNumaMigPartitionSizeSubArrayToMinimumValue(pKernelMemorySystem, swizzRange.lo, swizzRange.hi);
         }
@@ -719,6 +806,13 @@ kmemsysGetMIGGPUInstanceMemConfigFromSwizzId_IMPL
 )
 {
     NV_ASSERT_OR_RETURN(swizzId < KMIGMGR_MAX_GPU_SWIZZID, NV_ERR_INVALID_ARGUMENT);
+
+    if (IS_VIRTUAL(pGpu))
+    {
+        // VMMU Segment details are populated on Host and not Guest.
+        return NV_ERR_NOT_SUPPORTED;
+    }
+
     // MODS makes a control call to describe GPU instances before this is populated. Return invalid data anyways
     NV_ASSERT_OR_RETURN(pKernelMemorySystem->gpuInstanceMemConfig[swizzId].bInitialized, NV_ERR_INVALID_STATE);
 
@@ -802,50 +896,46 @@ kmemsysSetupCoherentCpuLink_IMPL
     KernelBus     *pKernelBus     = GPU_GET_KERNEL_BUS(pGpu);
     MemoryManager *pMemoryManager = GPU_GET_MEMORY_MANAGER(pGpu);
     NvU64          numaOnlineSize = 0;
-    NvU64          fbSize         = (pMemoryManager->Ram.fbTotalMemSizeMb << 20);
+    NvU64          coherentCpuFbSize = 0;
     NvU32          data32;
-    NvBool         bCpuMapping    = NV_TRUE; // Default enable
     NvS32          numaNodeId     = NV0000_CTRL_NO_NUMA_NODE;
     NvU64          memblockSize   = 0;
-    NvU64          numaOnlineBase = 0;
     NvU64          rsvdFastSize   = 0;
     NvU64          rsvdSlowSize   = 0;
     NvU64          rsvdISOSize    = 0;
     NvU64          totalRsvdBytes = 0;
 
+    // Parse regkey here
+    if ((osReadRegistryDword(pGpu,
+                             NV_REG_STR_RM_FORCE_BAR_PATH, &data32) == NV_OK) &&
+        (data32 == 1))
     {
-        NV_ASSERT_OK_OR_RETURN(kmemsysGetFbNumaInfo_HAL(pGpu, pKernelMemorySystem,
-                                                        &pKernelMemorySystem->coherentCpuFbBase,
-                                                        &numaNodeId));
-        if (pKernelMemorySystem->coherentCpuFbBase != 0)
-        {
-            if (gpuIsSelfHosted(pGpu))
-            {
-                //
-                // For self-hosted, coherentCpuFbEnd is only till the FB size
-                // end and NOT till the FB AMAP end since self-hosted doesn't
-                // support indirect peer and requires GPU nvlink for peer.
-                //
-                pKernelMemorySystem->coherentCpuFbEnd = pKernelMemorySystem->coherentCpuFbBase + fbSize;
-            }
-            else
-            {
-                RM_API *pRmApi = GPU_GET_PHYSICAL_RMAPI(pGpu);
-                NV2080_CTRL_INTERNAL_GET_COHERENT_FB_APERTURE_SIZE_PARAMS params = {0};
+        NV_PRINTF(LEVEL_ERROR,
+                  "Force disabling NVLINK/C2C mappings through regkey.\n");
+        return NV_OK;
+    }
 
-                NV_ASSERT_OK_OR_RETURN(pRmApi->Control(pRmApi,
-                                                       pGpu->hInternalClient,
-                                                       pGpu->hInternalSubdevice,
-                                                       NV2080_CTRL_CMD_INTERNAL_GET_COHERENT_FB_APERTURE_SIZE,
-                                                       &params,
-                                                       sizeof(NV2080_CTRL_INTERNAL_GET_COHERENT_FB_APERTURE_SIZE_PARAMS)));
-                //
-                // Indirect peer(uses P9 to reach other GV100) in P9+GV100 requires coherentCpuFbEnd to
-                // also include the entire FB AMAP range even when FB size is less than the FB AMAP size.
-                //
-                pKernelMemorySystem->coherentCpuFbEnd = pKernelMemorySystem->coherentCpuFbBase +
-                                                        params.coherentFbApertureSize;
-            }
+    NV_ASSERT_OK_OR_RETURN(kmemsysGetFbNumaInfo_HAL(pGpu, pKernelMemorySystem,
+                                                    &pKernelMemorySystem->coherentCpuFbBase,
+                                                    &coherentCpuFbSize,
+                                                    &pKernelMemorySystem->coherentRsvdFbBase,
+                                                    &numaNodeId));
+
+    if (coherentCpuFbSize == 0)
+    {
+        //
+        // For passthrough, OS layer may be unable to provide the CPU-coherent size of FB (as it
+        // won't be available in ACPI DSD), so try to determine it from VBIOS scratch.
+        //
+        if (IS_PASSTHRU(pGpu))
+        {
+            NV_ASSERT_OK_OR_RETURN(kmemsysReadHdmTopFromVbios_HAL(pGpu, pKernelMemorySystem, &coherentCpuFbSize));
+        }
+
+        // Otherwise, if we are still unable to determine the CPU-coherent size, assume whole FB is CPU-coherent
+        if (coherentCpuFbSize == 0)
+        {
+            coherentCpuFbSize = (pMemoryManager->Ram.fbTotalMemSizeMb << 20);
         }
     }
 
@@ -856,22 +946,36 @@ kmemsysSetupCoherentCpuLink_IMPL
         NV_PRINTF(LEVEL_ERROR, "Override GPU NUMA node ID %d!\n", numaNodeId);
     }
 
-    // Parse regkey here
-    if ((osReadRegistryDword(pGpu,
-                             NV_REG_STR_RM_FORCE_BAR_PATH, &data32) == NV_OK) &&
-        (data32 == 1))
+    //
+    // EnableUserNUMAManagement is the NV_REG_ENABLE_USER_NUMA_MANAGEMENT is a unix-layer regkey
+    // If NUMA onlining is explicitly disabled by regkey, then don't check for valid node.
+    // The rest of the code for mapping over coherent links will use osNumaOnliningEnabled
+    // to check whether we should use the NUMA path, but in this spot, we need to check for
+    // the regkey itself because we still need to sanity check we have a valid node when
+    // osNumaOnliningEnabled is supposed to work.
+    //
+    if (!(osReadRegistryDword(pGpu, "EnableUserNUMAManagement", &data32) == NV_OK && (data32 == 0)))
     {
-        NV_PRINTF(LEVEL_ERROR,
-                  "Force disabling NVLINK/C2C mappings through regkey.\n");
-
-        bCpuMapping = NV_FALSE;
+        if ((IS_SILICON(pGpu) && (numaNodeId == NV0000_CTRL_NO_NUMA_NODE) && !hypervisorIsVgxHyper()))
+        {
+            //
+            // Do not fail for vGPU host as the device memory is not added to
+            // the kernel and it is expected for node id to not be set.
+            // Do not fail if NUMA onlining is explicitly disabled either
+            //
+            NV_PRINTF(LEVEL_ERROR, "Failed to get NUMA node id for GPU memory\n");
+            return NV_ERR_INVALID_STATE;
+        }
     }
 
-    if ((pKernelMemorySystem->coherentCpuFbBase == 0) || !bCpuMapping)
+    if (pKernelMemorySystem->coherentCpuFbBase == 0)
     {
-        return NV_OK;
+        NV_PRINTF(LEVEL_ERROR, "Failed to get coherent GPU memory base address\n");
+        return NV_ERR_INVALID_STATE;
     }
 
+    pKernelMemorySystem->coherentCpuFbEnd = pKernelMemorySystem->coherentCpuFbBase + coherentCpuFbSize;
+
     NV_ASSERT_OK_OR_RETURN(osNumaMemblockSize(&memblockSize));
 
     memmgrCalcReservedFbSpaceHal_HAL(pGpu, pMemoryManager, &rsvdFastSize, &rsvdSlowSize, &rsvdISOSize);
@@ -881,7 +985,7 @@ kmemsysSetupCoherentCpuLink_IMPL
     // kernel after accounting for different reserved memory requirements.
     //
     // Align rsvd memory to 64K granularity.
-    // TODO : rsvdMemorySize is not finalized at this point of time in 
+    // TODO : rsvdMemorySize is not finalized at this point of time in
     // GH180, currently rsvdMemorySize is not increasing after this
     // point. This needs to be fixed.
     //
@@ -889,34 +993,59 @@ kmemsysSetupCoherentCpuLink_IMPL
     totalRsvdBytes += (rsvdFastSize + rsvdSlowSize + rsvdISOSize);
     totalRsvdBytes += pMemoryManager->Ram.reservedMemSize;
 
+    // For SRIOV guest, take into account FB tax paid on host side for each VF
+    // This FB tax is non zero only for SRIOV guest RM environment.
+    totalRsvdBytes += memmgrGetFbTaxSize_HAL(pGpu, pMemoryManager);
+
     //
     // TODO: make sure the onlineable memory is aligned to memblockSize
     // Currently, if we have leftover memory, it'll just be wasted because no
     // one can access it. If FB size itself is memblock size unaligned(because
     // of CBC and row remapper deductions), then the memory wastage is unavoidable.
     //
-    numaOnlineSize = NV_ALIGN_DOWN64(fbSize - totalRsvdBytes, memblockSize);
+    numaOnlineSize = KMEMSYS_FB_NUMA_ONLINE_SIZE(coherentCpuFbSize - totalRsvdBytes, memblockSize);
 
-    pKernelMemorySystem->numaOnlineBase   = numaOnlineBase;
-    pKernelMemorySystem->numaOnlineSize   = numaOnlineSize;
+    if (IS_PASSTHRU(pGpu) && pKernelMemorySystem->bBug3656943WAR)
+    {
+        // For passthrough case, reserved memory size is fixed as 1GB
+        NvU64 rsvdSize = 1 * 1024 * 1024 * 1024;
 
-    NV_PRINTF(LEVEL_INFO, "fbSize: 0x%llx NUMA reserved memory size: 0x%llx online memory size: 0x%llx\n",
-                  fbSize, totalRsvdBytes, numaOnlineSize);
-    //
-    // TODO: Bug 1945658: Soldier through on GPU memory add
-    // failure(which is often possible because of missing auto online
-    // setting) and instead check for failure on stateLoad.
-    // Any failure in StateInit results in gpuStateDestroy not getting called.
-    // kgspUnloadRm_IMPL from gpuStateDestroy also doesn't get called leaving
-    // GSP in unclean state and requiring GPU reset to recover from that.
-    //
-    // kmemsysNumaAddMemory_HAL by itself cannot be called from stateLoad
-    // because the memory mapping that follows this call site comes from linear
-    // kernel virtual address when memory is added to the kernel vs the
-    // VMALLOC_START region when memory is not added.
-    //
-    NV_ASSERT_OK(kmemsysNumaAddMemory_HAL(pGpu, pKernelMemorySystem, 0, 0,
-                                          numaOnlineSize, &numaNodeId));
+        NV_ASSERT_OR_RETURN(rsvdSize >= totalRsvdBytes, NV_ERR_INVALID_STATE);
+        totalRsvdBytes = rsvdSize;
+        //
+        // Aligning to hardcoded 512MB size as both host and guest need to use
+        // the same alignment irrespective of the kernel page size. 512MB size
+        // works for both 4K and 64K page size kernels but more memory is
+        // wasted being part of non onlined region which can't be avoided
+        // per the design.
+        //
+        numaOnlineSize = KMEMSYS_FB_NUMA_ONLINE_SIZE(coherentCpuFbSize - totalRsvdBytes, 512 * 1024 * 1024);
+    }
+
+    NV_PRINTF(LEVEL_INFO,
+              "coherentCpuFbSize: 0x%llx NUMA reserved memory size: 0x%llx online memory size: 0x%llx\n",
+              coherentCpuFbSize, totalRsvdBytes, numaOnlineSize);
+
+    if (osNumaOnliningEnabled(pGpu->pOsGpuInfo))
+    {
+        pKernelMemorySystem->numaOnlineBase   = KMEMSYS_FB_NUMA_ONLINE_BASE;
+        pKernelMemorySystem->numaOnlineSize   = numaOnlineSize;
+        //
+        // TODO: Bug 1945658: Soldier through on GPU memory add
+        // failure(which is often possible because of missing auto online
+        // setting) and instead check for failure on stateLoad.
+        // Any failure in StateInit results in gpuStateDestroy not getting called.
+        // kgspUnloadRm_IMPL from gpuStateDestroy also doesn't get called leaving
+        // GSP in unclean state and requiring GPU reset to recover from that.
+        //
+        // kmemsysNumaAddMemory_HAL by itself cannot be called from stateLoad
+        // because the memory mapping that follows this call site comes from linear
+        // kernel virtual address when memory is added to the kernel vs the
+        // VMALLOC_START region when memory is not added.
+        //
+        NV_ASSERT_OK(kmemsysNumaAddMemory_HAL(pGpu, pKernelMemorySystem, 0, 0,
+                                              numaOnlineSize, &numaNodeId));
+    }
     pGpu->numaNodeId = numaNodeId;
 
     NV_ASSERT_OK_OR_RETURN(kbusCreateCoherentCpuMapping_HAL(pGpu, pKernelBus, numaOnlineSize, bFlush));
@@ -924,8 +1053,6 @@ kmemsysSetupCoherentCpuLink_IMPL
     // Switch the toggle for coherent link mapping only if migration is successful
     pGpu->setProperty(pGpu, PDB_PROP_GPU_COHERENT_CPU_MAPPING, NV_TRUE);
 
-    NV_ASSERT_OK_OR_RETURN(kbusVerifyCoherentLink_HAL(pGpu, pKernelBus));
-
     return NV_OK;
 }
 
@@ -991,3 +1118,100 @@ kmemsysGetUsableFbSize_KERNEL
     return kmemsysReadUsableFbSize_HAL(pGpu, pKernelMemorySystem, pFbSize);
 }
 
+NV_STATUS
+kmemsysStateLoad_IMPL(OBJGPU *pGpu, KernelMemorySystem *pKernelMemorySystem, NvU32 flags)
+{
+    NV_STATUS status = NV_OK;
+
+    if (IS_GSP_CLIENT(pGpu) || IS_VIRTUAL(pGpu))
+    {
+        if ((flags & GPU_STATE_FLAGS_PRESERVING) &&
+            (flags & GPU_STATE_FLAGS_PM_TRANSITION) &&
+            !(flags & GPU_STATE_FLAGS_GC6_TRANSITION))
+        {
+            MemoryManager *pMemoryManager = GPU_GET_MEMORY_MANAGER(pGpu);
+
+            NV_ASSERT_OK_OR_ELSE(status,
+                                 memmgrRestorePowerMgmtState(pGpu, pMemoryManager),
+                                 memmgrFreeFbsrMemory(pGpu, pMemoryManager));
+        }
+    }
+
+    return status;
+}
+
+NV_STATUS
+kmemsysStateUnload_IMPL(OBJGPU *pGpu, KernelMemorySystem *pKernelMemorySystem, NvU32 flags)
+{
+    NV_STATUS status = NV_OK;
+
+    if (IS_GSP_CLIENT(pGpu) || IS_VIRTUAL(pGpu))
+    {
+        if ((flags & GPU_STATE_FLAGS_PRESERVING) &&
+            (flags & GPU_STATE_FLAGS_PM_TRANSITION) &&
+            !(flags & GPU_STATE_FLAGS_GC6_TRANSITION))
+        {
+            MemoryManager *pMemoryManager = GPU_GET_MEMORY_MANAGER(pGpu);
+
+            NV_ASSERT_OK_OR_ELSE(status,
+                                 memmgrSavePowerMgmtState(pGpu, pMemoryManager),
+                                 memmgrFreeFbsrMemory(pGpu, pMemoryManager));
+        }
+    }
+
+    return status;
+}
+
+/*!
+ * Called after the video memory heap is created
+ */
+NV_STATUS
+kmemsysPostHeapCreate_KERNEL
+(
+    POBJGPU               pGpu,
+    KernelMemorySystem   *pKernelMemorySystem
+)
+{
+    if (pGpu->getProperty(pGpu, PDB_PROP_GPU_COHERENT_CPU_MAPPING))
+    {
+        NV_ASSERT_OK_OR_RETURN(kbusVerifyCoherentLink_HAL(pGpu, GPU_GET_KERNEL_BUS(pGpu)));
+    }
+
+    return NV_OK;
+}
+
+NV_STATUS
+kmemsysGetCpuCoherentFbRange_IMPL
+(
+    OBJGPU *pGpu,
+    KernelMemorySystem *pKernelMemorySystem,
+    NvU64 *pCpuCoherentFbBase,
+    NvU64 *pCpuCoherentFbEnd
+)
+{
+    if ((pCpuCoherentFbBase == NULL) || (pCpuCoherentFbEnd == NULL))
+    {
+        return NV_ERR_INVALID_ARGUMENT;
+    }
+
+    // CPU-coherent FB memory only applicable to Self-Hosted platforms
+    if (!gpuIsSelfHosted(pGpu))
+    {
+        return NV_ERR_NOT_SUPPORTED;
+    }
+
+    if (pKernelMemorySystem->coherentCpuFbBase != 0)
+    {
+        *pCpuCoherentFbBase = pKernelMemorySystem->coherentCpuFbBase;
+        *pCpuCoherentFbEnd = pKernelMemorySystem->coherentCpuFbEnd;
+    }
+    else
+    {
+        *pCpuCoherentFbBase = 0;
+        *pCpuCoherentFbEnd = 0;
+    }
+
+    return NV_OK;
+}
+
+
